/* Creates "the-fifo" in the current directory, if it
   does not exist already. */

#include <stdio.h>
#include <fcntl.h>
#include <sys/event.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>

static void fail(const char *s);
static void watch_one_shot(int kq, int fd, int filt);
static int fill_up(int w);
static int drain(int r);
static void prepare_fifo(const char *fn);

int main()
{
  struct kevent kev[1];
  int r, w, kq;
  const char *fn = "the-fifo";
  int v;

  prepare_fifo(fn);

  kq = kqueue();
  if (kq < 0) fail("kqueue failed");

  r = open(fn, O_RDONLY | O_APPEND | O_NONBLOCK);
  if (r < 0) fail("open for read failed");

  w = open(fn, O_WRONLY | O_APPEND | O_NONBLOCK);
  if (w < 0) fail("open for write failed");

  watch_one_shot(kq, w, EVFILT_WRITE);
  watch_one_shot(kq, r, EVFILT_READ);

  while (1) {
    v = kevent(kq, NULL, 0, kev, 1, NULL);
    if (v < 0) fail("kevent failed");

    if (kev[0].ident == w) {
      printf("write ready\n");
      v = fill_up(w);
      printf("wrote %d\n", v);
      watch_one_shot(kq, w, EVFILT_WRITE);
    } else if (kev[0].ident == r) {
      printf("read ready\n");
      v = drain(r);
      printf("read %d\n", v);
      watch_one_shot(kq, r, EVFILT_READ);
      /* Re-watching output here works around the bug: */
      /* watch_one_shot(kq, w, EVFILT_WRITE); */
    }
  }

  return 0;
}

void fail(const char *s)
{
  printf("%s\n", s);
  abort();
}

void watch_one_shot(int kq, int fd, int filt)
{
  struct kevent kev[1];
  struct timespec timeout = {0, 0};
  int v, pos;

  if (kq < 0) return;

  if (filt == EVFILT_READ)
    printf("watch read\n");
  else if (filt == EVFILT_WRITE)
    printf("watch write\n");

  EV_SET(kev, fd, filt, EV_ADD | EV_ONESHOT, 0, 0, NULL);

  v = kevent(kq, kev, 1, NULL, 0, &timeout);
  if (v) fail("kevent failed");
}

int fill_up(int w)
{
  int v, count = 0, amt = 5;

  while (1) {
    v = write(w, "hello", amt);
    if (v < 0) {
      if (errno == EAGAIN) {
        if (amt == 1)
          return count;
        else
          amt -= 1;
      } else
        fail("write failed");
    } else
      count += v;
  }
}

int drain(int r)
{
  int v, count = 0;
  char buf[256];

  while (1) {
    v = read(r, buf, 256);
    if (v < 0) {
      if (errno == EAGAIN)
        return count;
      fail("read failed");
    } else
      count += v;
  }
}

void prepare_fifo(const char *fn)
{
  struct stat s_buf;
  int v;

  /* Create or check "the-fifo": */
  v = stat(fn, &s_buf);
  if (!v) {
    if (!S_ISFIFO(s_buf.st_mode))
      fail("not a fifo");
  } else if (errno == ENOENT) {
    v = mkfifo(fn, 0777);
    if (v < 0) fail("mkfifo failed");
  } else
    fail("stat failed");
}
